from copy import deepcopy
from inflection import underscore
from kant.eventstore.stream import EventStream
from kant.datamapper.base import ModelMeta, FieldMapping
from kant.datamapper.fields import *  # NOQA


class ProjectionManager:
    def __init__(self):
        self._adapters = set()

    def bind(self, adapter):
        self._adapters.add(adapter)

    async def notify_create(self, *args, **kwargs):
        for adapter in self._adapters:
            await adapter.handle_create(*args, **kwargs)

    async def notify_update(self, *args, **kwargs):
        for adapter in self._adapters:
            await adapter.handle_update(*args, **kwargs)


class Projection(FieldMapping, metaclass=ModelMeta):
    def fetch_events(self, eventstream):
        for event in eventstream:
            self.when(event)

    def when(self, event):
        event_name = underscore(event.__class__.__name__)
        method_name = 'when_{0}'.format(event_name)
        try:
            method = getattr(self, method_name)
            method(event)
        except AttributeError:
            pass


class ProjectionRouter:
    def __init__(self):
        self._projections = {}
        self._models = {}

    def add(self, keyspace, model, projection):
        self._models[keyspace] = model
        self._projections[keyspace] = projection

    def get_projection(self, keyspace, eventstream):
        if keyspace not in self._projections:
            raise ProjectionDoesNotExist(keyspace)
        Projection = self._projections[keyspace]
        projection = Projection()
        projection.fetch_events(eventstream)
        return projection

    def get_model(self, keyspace):
        if keyspace not in self._models:
            raise ProjectionDoesNotExist(keyspace)
        return self._models[keyspace]



from .base import ModelMeta, FieldMapping


class SchemaModel(FieldMapping, metaclass=ModelMeta):

    @classmethod
    def make(self, obj, cls):
        Event = cls
        json_columns = {}
        for name, field in Event.concrete_fields.items():
            json_column = field.json_column or name
            json_columns[json_column] = name
        args = {json_columns[name]: value for name, value in obj.items()}
        return Event(**args)

    def decode(self):
        event = {key: value for key, value in self.serializeditems()}
        return event



from copy import deepcopy
from async_generator import yield_, async_generator
from inflection import underscore
from kant.datamapper.base import ModelMeta, FieldMapping
from kant.datamapper.fields import *  # NOQA
from kant.eventstore import EventStream, get_connection
from .exceptions import AggregateError


class Manager:
    def __init__(self, model, keyspace, using=None):
        self._model = model
        self.using = using
        self.keyspace = keyspace

    @property
    def _conn(self):
        if self.using is None:
            return get_connection()
        return self.using

    async def save(self, aggregate_id, events, notify_save):
        async with self._conn.open(self.keyspace) as eventstore:
            await eventstore.append_to_stream(aggregate_id, events, notify_save)

    async def get(self, aggregate_id):
        async with self._conn.open(self.keyspace) as eventstore:
            stream = await eventstore.get_stream(aggregate_id)
            return self._model.from_stream(stream)

    @async_generator
    async def all(self):
        async with self._conn.open(self.keyspace) as eventstore:
            async for stream in eventstore.all_streams():
                await yield_(self._model.from_stream(stream))

    async def get_stream(self, aggregate_id):
        async with self._conn.open(self.keyspace) as eventstore:
            return await eventstore.get_stream(aggregate_id)


class AggregateMeta(ModelMeta):
    def __new__(mcs, class_name, bases, attrs):
        cls = ModelMeta.__new__(mcs, class_name, bases, attrs)
        if '__keyspace__' in attrs.keys():
            cls.objects = Manager(model=cls, keyspace=attrs['__keyspace__'])
        return cls


class Aggregate(FieldMapping, metaclass=AggregateMeta):
    def __init__(self):
        super().__init__()
        self._all_events = EventStream()
        self._events = EventStream()
        self._stored_events = EventStream()

    def all_events(self):
        return self._all_events

    def stored_events(self):
        return self._stored_events

    def get_events(self):
        return self._events

    def notify_save(self, new_version):
        self._events.clear()
        self._events.initial_version = new_version
        self._all_events.initial_version = new_version
        self._stored_events.initial_version = new_version

    def fetch_events(self, events: EventStream):
        self._stored_events = deepcopy(events)
        self._all_events = deepcopy(self._stored_events)
        self._events.initial_version = events.initial_version
        for event in events:
            self.dispatch(event, flush=False)

    def apply(self, event):
        event_name = underscore(event.__class__.__name__)
        method_name = 'apply_{0}'.format(event_name)
        try:
            method = getattr(self, method_name)
            method(event)
        except AggregateError:
            msg = "The command for '{}' is not defined".format(event.__class__.__name__)
            raise CommandError(msg)

    def dispatch(self, events, flush=True):
        if isinstance(events, list):
            received_events = list(events)
        else:
            received_events = [events]
        for event in received_events:
            self.apply(event)
            if flush:
                self._events.add(event)
                self._all_events.add(event)

    @property
    def current_version(self):
        return self._all_events.current_version

    @property
    def version(self):
        return self._all_events.initial_version

    @classmethod
    def from_stream(cls, stream):
        self = cls()
        self.fetch_events(stream)
        return self

    async def save(self):
        return await self.objects.save(self.get_pk(), self.get_events(), self.notify_save)

    async def refresh_from_db(self):
        stream = await self.objects.get_stream(self.get_pk())
        self.fetch_events(stream)
        return self

    def get_pk(self):
        primary_keys = list(self.primary_keys().values())
        if not primary_keys:
            msg = "Nothing primary key defined for '{}'".format(self.__class__.__name__)
            raise AggregateError(msg)
        elif len(primary_keys) > 1:
            msg = "Many primary keys defined for '{}'".format(self.__class__.__name__)
            raise AggregateError(msg)
        return primary_keys[0]



from sqlalchemy import and_, literal_column
from .exceptions import ProjectionError, ProjectionDoesNotExist


class SQLAlchemyProjectionAdapter:
    def __init__(self, saconnection, router):
        self.saconnection = saconnection
        self.router = router

    async def handle_create(self, keyspace, steam, eventstream):
        projection = self.router.get_projection(keyspace, eventstream)
        stmt = self.router.get_model(keyspace).insert().values(**projection.decode())
        await self.saconnection.execute(stmt)

    async def handle_update(self, keyspace, steam, eventstream):
        projection = self.router.get_projection(keyspace, eventstream)
        if not projection.primary_keys():
            msg = "'{}' not have primary keys".format(projection.__class__.__name__)
            raise ProjectionError(msg)
        stmt = self.router.get_model(keyspace).update().values(
            **projection.decode()
        ).where(
            and_(
                literal_column(field) == value for field, value in projection.primary_keys().items()
            )
        )
        await self.saconnection.execute(stmt)



from abc import ABCMeta
from datetime import datetime
from decimal import Decimal
from dateutil import parser as dateutil_parser
from cuid import cuid


class Field(metaclass=ABCMeta):
    def __init__(self, default=None, json_column=None, primary_key=False, *args, **kwargs):
        self.primary_key = primary_key
        self.default = default
        self.json_column = json_column

    def encode(self, value):
        return value

    def parse(self, value):
        return value

    def default_value(self):
        if callable(self.default):
            return self.default()
        return self.default


class CUIDField(Field):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.primary_key:
            self.default = cuid

    def encode(self, value):
        return str(value)

    def parse(self, value):
        return value


class DecimalField(Field):
    def encode(self, value):
        return float(value)

    def parse(self, value):
        return Decimal(value)


class IntegerField(Field):
    def encode(self, value):
        return int(value)

    def parse(self, value):
        return int(value)


class DateTimeField(Field):
    def __init__(self, auto_now=False, *args, **kwargs):
        self.auto_now = auto_now
        super().__init__(*args, **kwargs)
        if self.auto_now:
            self.default = lambda: datetime.now()

    def encode(self, value):
        """
        >>> field = DateTimeField()
        >>> create_at = datetime(2009, 5, 28, 16, 15)
        >>> field.encode(create_at)
        '2009-05-28T16:15:00'
        """
        return value.isoformat()

    def parse(self, value):
        """
        >>> field = DateTimeField()
        >>> field.parse('2009-05-28T16:15:00')
        datetime.datetime(2009, 5, 28, 16, 15)
        >>> create_at = datetime(2009, 5, 28, 16, 15)
        >>> field.parse(create_at)
        datetime.datetime(2009, 5, 28, 16, 15)
        """
        if isinstance(value, str):
            return dateutil_parser.parse(value)
        if not isinstance(value, datetime):
            raise TypeError('expected string or datetime object')
        return value


class CharField(Field):
    def encode(self, value):
        """
        >>> field = CharField()
        >>> field.encode(123)
        '123'
        """
        return str(value)

    def parse(self, value):
        """
        >>> field = CharField()
        >>> field.parse(123)
        '123'
        """
        return str(value)


class BooleanField(Field):
    def encode(self, value):
        """
        >>> field = BooleanField()
        >>> field.encode(True)
        'true'
        >>> field.encode(False)
        'false'
        """
        if value:
            return 'true'
        return 'false'

    def parse(self, value):
        """
        >>> field = BooleanField()
        >>> field.parse(True)
        True
        >>> field.parse(False)
        False
        >>> field.parse('true')
        True
        >>> field.parse('false')
        False
        """
        if isinstance(value, str) and value.strip() == 'false':
            return False
        return bool(value)


class SchemaField(Field):
    def __init__(self, to, *args, **kwargs):
        self.to = to
        super().__init__(*args, **kwargs)

    def encode(self, value):
        """
        >>> account = AccountSchemaModel(balance=25.5)
        >>> field = SchemaField(to=AccountSchemaModel)
        >>> field.encode(account)
        {'balance': 25.5}
        """
        return value.decode()

    def parse(self, value):
        """
        >>> account = AccountSchemaModel(balance=25.5)
        >>> field = SchemaField(to=AccountSchemaModel)
        >>> field.parse(account)
        <AccountSchemaModel: {'balance': Decimal('25.5')}>
        >>> field.parse({"balance": 30 })
        <AccountSchemaModel: {'balance': Decimal('30')}>
        """
        if isinstance(value, self.to):
            return value
        return self.to().make(value, cls=self.to)



from abc import ABCMeta
from collections import MutableMapping
import json
from .fields import Field
from .exceptions import FieldError


class ModelMeta(ABCMeta):
    def __new__(mcs, class_name, bases, attrs):
        concrete_fields = {}
        new_attrs = {}
        for name, value in attrs.items():
            if isinstance(value, Field):
                concrete_fields[name] = value
            else:
                new_attrs[name] = value

        cls = type.__new__(mcs, class_name, bases, new_attrs)
        cls.concrete_fields = cls.concrete_fields.copy()
        cls.concrete_fields.update(concrete_fields)
        return cls


class FieldMapping(MutableMapping):
    concrete_fields = {}

    def __init__(self, *args, **kwargs):
        self._values = {}
        initial = {}
        for name, value in self.concrete_fields.items():
            if value.default is not None:
                initial[name] = value.default_value()
        initial.update(dict(*args, **kwargs))
        if args or kwargs:  # avoid creating dict for most common case
            for name, value in initial.items():
                self[name] = value

    def __getitem__(self, key):
        return self._values[key]

    def __setitem__(self, key, value):
        if key in self.concrete_fields:
            try:
                self._values[key] = self.concrete_fields[key].parse(value)
            except TypeError as e:
                msg = ("The value '{value}' is invalid. "
                       "The field '{key}' {exception}").format(
                    value=repr(value),
                    field=key,
                    exception=str(e),
                )
                raise TypeError(msg)
        else:
            msg = '{class_name} does not support field: {field}'.format(
                class_name=self.__class__.__name__,
                field=key,
            )
            raise KeyError(msg)

    def __delitem__(self, key):
        del self._values[key]

    def __getattr__(self, name):
        if name not in self.concrete_fields:
            return super().__getattribute__(name)
        return self._values[name]

    def __setattr__(self, name, value):
        if name not in self.concrete_fields:
            return super().__setattr__(name, value)
        try:
            if value is not None:
                self._values[name] = self.concrete_fields[name].parse(value)
        except KeyError:
            msg = "The field '{0}' is not defined in '{1}'.".format(name, self.__class__.__name__)
            raise FieldError(msg)

    def keys(self):
        return self._values.keys()

    def __repr__(self):
        return '<{0}: {1}>'.format(self.__class__.__name__, self._values)

    def copy(self):
        return self.__class__(self)

    def __len__(self):
        return len(self._values)

    def __iter__(self):
        return iter(self._values)

    def __hash__(self):
        return id(self)

    def serializeditems(self):
        for name, value in self._values.items():
            field_name = self.concrete_fields[name].json_column or name
            field_value = self.concrete_fields[name].encode(value)
            yield (field_name, field_value)

    def decode(self):
        return {key: value for key, value in self.serializeditems()}

    def json(self, only=None):
        data = self.decode()
        if only is not None:
            data = {key: value for key, value in data.items() if key in only}
        return json.dumps(data, sort_keys=True)

    def primary_keys(self):
        primary_keys = {}
        for name, value in self._values.items():
            field = self.concrete_fields[name]
            field_name = field.json_column or name
            field_value = field.encode(value)
            if field.primary_key:
                primary_keys[field_name] = field_value
        return primary_keys



class VersionError(Exception):
    pass


class StreamExists(VersionError):
    def __init__(self, event, *args, **kwargs):
        message = "The EventStream is not empty. The Event '{}' cannot be added.".format(event.__class__.__name__)
        super().__init__(message, *args, **kwargs)


class DependencyDoesNotExist(VersionError):
    def __init__(self, event, dependencies, *args, **kwargs):
        message = "The Event '{}' expected {}.".format(
            event.__class__.__name__,
            dependencies,
        )
        super().__init__(message, *args, **kwargs)


class DatabaseError(Exception):
    pass


class IntegrityError(DatabaseError):
    pass


class StreamDoesNotExist(DatabaseError):
    pass



from collections import namedtuple
from asyncio_extras.contextmanager import async_contextmanager
from async_generator import yield_, async_generator
import aiopg
from kant.projections import ProjectionManager
from ..exceptions import (StreamDoesNotExist, VersionError)
from ..stream import EventStream


class EventStoreConnection:
    def __init__(self):
        self.projections = ProjectionManager()

    @classmethod
    async def create(cls, settings):
        self = EventStoreConnection()
        self.settings = settings
        self.pool = settings.get('pool')
        if self.pool is None:
            self.pool = await aiopg.connect(
                user=settings.get('user'),
                password=settings.get('password'),
                database=settings.get('database'),
                host=settings.get('host'),
            )
        return self

    async def create_keyspace(self, keyspace):
        stmt = """
        CREATE TABLE IF NOT EXISTS {keyspace} (
            id varchar(255),
            data jsonb NOT NULL,
            created_at timestamp NOT NULL,
            updated_at timestamp NOT NULL,
            version bigserial NOT NULL
        )
        """.format(keyspace=keyspace)
        async with self.pool.cursor() as cursor:
            await cursor.execute(stmt)

    async def drop_keyspace(self, keyspace):
        stmt = """
        DROP TABLE {keyspace}
        """.format(keyspace=keyspace)
        async with self.pool.cursor() as cursor:
            await cursor.execute(stmt)

    @async_contextmanager
    async def open(self, keyspace):
        async with self.pool.cursor() as cursor:
            await yield_(EventStore(cursor, keyspace, self.projections))


class EventStore:
    def __init__(self, cursor, keyspace, projections):
        self.cursor = cursor
        self.keyspace = keyspace
        self.projections = projections

    async def get_stream(self, stream: str, start: int = 0, backward: bool = False):
        stmt_select = """
        SELECT {keyspace}.data
        FROM {keyspace} WHERE {keyspace}.id = %(id)s
        """.format(keyspace=self.keyspace)
        if backward:
            stmt_select += " AND CAST(data ? '$version' AS INTEGER) <= %(version)s"
        else:
            stmt_select += " AND CAST(data ? '$version' AS INTEGER) >= %(version)s"
        await self.cursor.execute(stmt_select, {
            'id': str(stream),
            'version': start,
        })
        eventstore_stream = await self.cursor.fetchone()
        if not eventstore_stream:
            raise StreamDoesNotExist(stream)
        return EventStream.make(eventstore_stream[0])

    @async_generator
    async def all_streams(self, start: int = 0, end: int = -1):
        stmt_select = """
        SELECT {keyspace}.data
        FROM {keyspace}
        ORDER BY id
        """.format(keyspace=self.keyspace)
        if start > 0:
            stmt_select += " OFFSET {}".format(start)
        if end > -1:
            stmt_select += " LIMIT {}".format(end)
        await self.cursor.execute(stmt_select)
        eventstore = await self.cursor.fetchall()
        for stream in eventstore:
            await yield_(EventStream.make(stream[0]))

    async def append_to_stream(self, stream: str, eventstream: EventStream, on_save=None):
        try:
            stored_eventstream = await self.get_stream(stream)

            if stored_eventstream.current_version > eventstream.initial_version:
                message = "The version '{0}' was expected in '{1}'".format(
                    eventstream.initial_version,
                    stored_eventstream,
                )
                raise VersionError(message)

            stored_eventstream += eventstream

            stmt_update = """
            UPDATE {keyspace} SET data=%(data)s, updated_at=NOW(), version=%(current_version)s
            WHERE id = %(id)s AND version = %(initial_version)s;
            """.format(keyspace=self.keyspace)
            await self.cursor.execute(stmt_update, {
                'id': str(stream),
                'initial_version': stored_eventstream.initial_version,
                'current_version': stored_eventstream.current_version,
                'data': stored_eventstream.json(),
            })
            if self.cursor.rowcount < 1:
                message = "The version '{0}' was expected in '{1}'".format(
                    eventstream.initial_version,
                    stored_eventstream,
                )
                raise VersionError(message)
            await self.projections.notify_update(self.keyspace, stream, stored_eventstream)
            if on_save is not None:
                on_save(stored_eventstream.current_version)
        except StreamDoesNotExist:
            stmt_insert = """
            INSERT INTO {keyspace} (id, version, data, created_at, updated_at)
            VALUES (%(id)s, %(version)s, %(data)s, NOW(), NOW())
            """.format(keyspace=self.keyspace)
            await self.cursor.execute(stmt_insert, {
                'id': str(stream),
                'version': eventstream.current_version,
                'data': eventstream.json(),
            })
            await self.projections.notify_create(self.keyspace, stream, eventstream)
            if on_save is not None:
                on_save(eventstream.current_version)



from .backends.aiopg import EventStoreConnection


_connection = None


async def connect(dsn=None, user=None, password=None, host=None, database=None, *, pool=None):
    global _connection
    settings = {
        'dsn': dsn,
        'user': user,
        'password': password,
        'host': host,
        'database': database,
        'pool': pool,
    }
    _connection = await EventStoreConnection.create(settings)
    return _connection


def get_connection():
    return _connection



from operator import attrgetter
import json
from kant.events import Event
from .exceptions import StreamExists, DependencyDoesNotExist


class EventStream:
    def __init__(self, events=None):
        self.initial_version = -1
        self.current_version = -1
        self._events = set()
        if events is not None:
            for event in list(events):
                self.initial_version += 1
                self.add(event)

    def __eq__(self, event_stream):
        return self.current_version == event_stream.current_version

    def __lt__(self, event_stream):
        return self.current_version < event_stream.current_version

    def __le__(self, event_stream):
        return self.current_version <= event_stream.current_version

    def __ne__(self, event_stream):
        return self.current_version != event_stream.current_version

    def __gt__(self, event_stream):
        return self.current_version > event_stream.current_version

    def __ge__(self, event_stream):
        return self.current_version >= event_stream.current_version

    def __len__(self):
        return len(self._events)

    def __add__(self, event_stream):
        for event in event_stream:
            self.add(event)
        return self

    def __iter__(self):
        return iter(sorted(self._events, key=attrgetter('version')))

    def __repr__(self):
        return str(list(self))

    def _valid_empty_stream(self, event):
        if event.__empty_stream__ and len(self._events) > 0:
            raise StreamExists(event)

    def _valid_dependencies(self, event):
        event_names = [event.__class__.__name__ for event in self._events]
        not_found = [event for event in event.__dependencies__ if event not in event_names]
        if len(not_found) > 0:
            raise DependencyDoesNotExist(event, not_found)

    def _conflict_resolution(self, event):
        self._valid_empty_stream(event)
        self._valid_dependencies(event)
        return event

    def exists(self):
        return self.initial_version != -1

    def clear(self):
        self._events = set()

    def add(self, event):
        if event not in self._events:
            event = self._conflict_resolution(event)
            self.current_version += 1
            event.version = self.current_version
            self._events.add(event)

    def decode(self):
        return [event.decode() for event in self._events]

    def json(self):
        return json.dumps(self.decode(), sort_keys=True)

    @classmethod
    def make(self, obj):
        return EventStream([Event.make(event) for event in obj])



from datetime import datetime
import json
import pytest
from kant import events


class FoundAdded(events.Event):
    amount = events.DecimalField()


class AccountSchemaModel(events.SchemaModel):
    balance = events.DecimalField()


@pytest.fixture(autouse=True)
def event_model_example(doctest_namespace):
    doctest_namespace['FoundAdded'] = FoundAdded
    doctest_namespace['json'] = json
    doctest_namespace['datetime'] = datetime
    doctest_namespace['AccountSchemaModel'] = AccountSchemaModel



from os import path
from setuptools import setup, find_packages


BASE_DIR = path.dirname(__file__)


with open(path.join(BASE_DIR, 'README.md')) as f:
    long_description = f.read()


setup(
    name='kant',
    version='2.0.2',
    description='The CQRS and Event Sourcing framework for Python',
    url='http://github.com/patrickporto/kant',
    author='Patrick Porto',
    author_email='patrick.s.porto@gmail.com',
    license='MIT',
    packages=find_packages(exclude=['conftest.py']),
    install_requires=[
        'python-dateutil',
        'inflection',
        'cuid.py',
        'async_generator',
        'asyncio_extras',
        'aiopg',
    ],
    long_description=long_description,
    setup_requires=[
        'pytest-runner',
    ],
    tests_require=[
        'pytest',
        'pytest-pycodestyle',
        'pytest-cov',
        'pytest-asyncio',
        'sqlalchemy',
    ],
    extras_require={
        'SQLAlchemy-Projections': ['sqlalchemy'],
    },
    zip_safe=False,
    classifiers=[
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: CPython',
    ],
    keywords='eventsourcing cqrs eventstore',
)



from abc import ABCMeta
from kant.datamapper.base import ModelMeta, FieldMapping
from kant.datamapper.fields import *  # NOQA
from kant.datamapper.models import *  # NOQA
from .exceptions import EventDoesNotExist, EventError


class Event(FieldMapping, metaclass=ModelMeta):
    EVENT_JSON_COLUMN = '$type'
    version = IntegerField(default=0, json_column='$version')
    __empty_stream__ = False
    __dependencies__ = []

    @classmethod
    def make(self, obj):
        if self.EVENT_JSON_COLUMN not in obj:
            msg = "'{}' is not defined in {}".format(self.EVENT_JSON_COLUMN, obj)
            raise EventError(msg)
        event_name = obj[self.EVENT_JSON_COLUMN]
        events = [Event for Event in self.__subclasses__() if Event.__name__ == event_name]
        try:
            Event = events[0]
        except IndexError:
            raise EventDoesNotExist()
        del obj[self.EVENT_JSON_COLUMN]
        json_columns = {}
        for name, field in Event.concrete_fields.items():
            json_column = field.json_column or name
            json_columns[json_column] = name
        args = {json_columns[name]: value for name, value in obj.items()}
        return Event(**args)

    def decode(self):
        event = {key: value for key, value in self.serializeditems()}
        event.update({
            self.EVENT_JSON_COLUMN: self.__class__.__name__,
        })
        return event
